// SPDX-FileCopyrightText: 2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "WlanCodeScannerServer.h"

// Qt includes
#include <QTcpSocket>
#include <QUrl>
#include <QUrlQuery>
#include <QTextStream>
#include <QTimer>

// C++ includes
#include <functional>

WlanCodeScannerServer::WlanCodeScannerServer(QObject *parent) : QTcpServer(parent)
{
    connect(this, &QTcpServer::newConnection, this, &WlanCodeScannerServer::processNewConnection);
}

void WlanCodeScannerServer::processNewConnection()
{
    auto *connection = nextPendingConnection();
    connect(connection, &QAbstractSocket::disconnected, connection, &QObject::deleteLater);
    connect(connection, &QTcpSocket::readyRead,
            this, std::bind(&WlanCodeScannerServer::processRequest, this, connection));
}

void WlanCodeScannerServer::processRequest(QTcpSocket *connection)
{
    if (! connection->isValid()) {
        // The connection has been closed in the meantime
        return;
    }

    // Read data from the socket
    QTextStream stream(connection);
    QString line;
    QVector<QString> lines;
    while (stream.readLineInto(&line)) {
        lines.append(line);
    }

    // Check if the header is complete
    if (lines.count() < 2 || ! lines.last().isEmpty()) {
        // It isn't, we have to wait for more
        return;
    }

    // The header is complete, so we search for the GET request we want to extract

    QString data;
    for (const auto &line : lines) {
        // The shortest possible parseable GET requst would be "GET /?content= HTTP/1.1", which is
        // 23 chars long. Also, the line obviously has to start with "GET" and end with "HTTP/1.1":
        if (line.count() >= 23
            && line.left(3) == QStringLiteral("GET")
            && line.right(8) == QStringLiteral("HTTP/1.1")) {

            // The length of "GET " is 4 chars, the length of " HTTP/1.1" is 9 chars. So, the
            // URL we want to extract starts at char 4 and is the total length - 13 chars long:
            const auto query = QUrlQuery(QUrl(line.mid(4, line.count() - 13)));
            data = query.queryItemValue(QStringLiteral("content"));
            data.replace(QLatin1Char('+'), QLatin1Char(' '));
            data = QString::fromUtf8(QByteArray::fromPercentEncoding(data.toUtf8()));
            break;
        }
    }

    // Answer the request

    QString message;
    if (! data.isEmpty()) {
        message = tr("Code empfangen!");
        Q_EMIT displayMessage(tr("Code empfangen! Der Datensatz ist:"));
        Q_EMIT displayMessage(data);
        // Announce the scan when we're completely done here
        QTimer::singleShot(0, this, std::bind(&WlanCodeScannerServer::codeReceived, this, data));

    } else {
        message = tr("Konnte den Code nicht verarbeiten :-(");
        Q_EMIT displayMessage(tr("Konnte eine Anfrage nicht verarbeiten. Die Anfrage war:"));
        Q_EMIT displayMessage(tr("--------------------------------------------------"));
        for (const auto &line : lines) {
            if (! line.isEmpty()) {
                Q_EMIT displayMessage(line);
            }
        }
        Q_EMIT displayMessage(tr("--------------------------------------------------"));
    }

    // Write the respective HTTP answer to the socket
    stream << "HTTP/1.1 200 OK\r\n"
           << "Content-Type: text/plain; charset=UTF-8\r\n"
           << "Content-Length: " << message.count() << "\r\n"
           << "\r\n"
           << message;

    connection->close();
}
